#!/usr/bin/env python3
"""Test if application ingress uses a valid certificate."""

import os
import shutil
import socket
import sys

import certifi
import pytest
import requests
from OpenSSL import SSL


@pytest.mark.resource
@pytest.mark.certs
def test_cert_validation(host, resource): # pylint: disable=too-many-statements
    """Checks for proper cluster certs from exposed services.
    Check is executed on the local provisioning machine.
    """

    def add_custom_cert_authorities(ca_file: str,
                                    custom_ca_files: list,
                                    dest_file: str =
                                    '/tmp/custom_ca_bundle.crt'):
        """Concatenates existing cert bundle with custom CAs."""

        destination = open(dest_file, 'wb')
        with open(dest_file, 'wb') as destination, open(ca_file, 'rb') as cert_auth:
            shutil.copyfileobj(cert_auth, destination)
            for custom_ca_file in custom_ca_files:
                with open(custom_ca_file, 'rb') as custom_ca:
                    shutil.copyfileobj(custom_ca, destination)


    def fetch_certs(domain: str, port: int = 443):
        """Fetches cert fom given domain."""

        sock = socket.socket()
        context = SSL.Context(SSL.TLSv1_2_METHOD)
        conn = SSL.Connection(context, sock)
        conn.set_tlsext_host_name(domain.encode('utf-8'))
        certs = []

        try:
            conn.connect((domain, port))
        except socket.gaierror as err:
            print('socket.gaierror: {0}'.format(str(err)))
            sys.exit(1)

        conn.do_handshake()
        try:
            certs = conn.get_peer_cert_chain()
        except SSL.Error as err:
            print('SSL Error: {0}'.format(str(err)))
            sys.exit(1)

        return certs


    def print_cert_info(certs: list):
        """Show CN and issuer CN of cert list."""

        for cert in certs:
            subject = cert.get_subject()
            common_name = subject.CN
            issuer = cert.get_issuer().CN
            print('CN: {0} (Issuer: {1})'.format(common_name, issuer))


    def valid_cert(domain: str, ca_file: str = '/tmp/custom_ca_bundle.crt',
                   app: str = "all"):
        """Validate cert of given domain against a ca_file bundle."""

        valid = False

        url = 'https://' + domain
        print('Validating cert from {0} ...'.format(url))
        certs = fetch_certs(domain)
        print_cert_info(certs)

        try:
            requests.get(url, verify=ca_file)
        except requests.exceptions.SSLError as ex:
            print('SSL Verification Error {}'.format(ex))
            for cert in certs:
                issuer = cert.get_issuer().CN
                if issuer == 'cert-manager.local':
                    print('Allowing exception for self-signed cert-mananger cert.')
                    valid = True
            return valid

        print('Successfully Verified SSL Cert.\n')
        return True


    app_subdomains = {
        'grafana': 'grafana',
        'nextcloud': 'files',
        'onlyoffice': 'office',
        # prometheus is only exposed if opted-in in the settings
        'prometheus': 'prometheus',
        'rocketchat': 'chat',
        'single-sign-on': 'sso',
        'wekan': 'wekan',
        'wordpress': 'www',
        'zulip': 'zulip',
    }

    if resource == 'all':
        apps = list(app_subdomains.keys())
    elif resource == 'kube-prometheus-stack':
        apps = ['grafana', 'prometheus']
    else:
        assert resource in app_subdomains, "Error: Unknown app: {}".format(app)
        apps = [resource]

    print('\n')
    for app_name in apps:
        print(f"{app_name}: ")
        # Use FQDN env var if set, else use hostname from ansible inventory
        domain = os.environ.get("FQDN")
        if domain:
            print("Using domain %s from FQDN environment variable." % domain)
        else:
            ansible_vars = host.ansible.get_variables()
            domain = ansible_vars["domain"]
            print("Using domain %s from ansible inventory." % domain)

        add_custom_cert_authorities(certifi.where(),
                                    ['pytest/le-staging-bundle.pem'])

        fqdn = app_subdomains[app_name] + '.' + domain
        assert valid_cert(domain=fqdn, app=resource)
